from pandioml.data.stream import Stream
import faker
from pandioml.data.record import JsonSchema, Record, String, Integer, Float
from decimal import Decimal


class Dataset(Stream):
    """
    A generator to produce an infinite amount of Person profiles using the Faker class.

    Each record is an instance of PersonProfile

    class PersonProfile(Record):
        first_name_nonbinary = String()
        last_name_nonbinary = String()
        address = String()
        ascii_email = String()
        user_name = String()
        language_name = String()
        latitude = Float()
        longitude = Float()
        ssn = String()
        ipv4 = String()
        unix_time = Integer()
    """
    dataset = None

    def __init__(self):
        self.dataset = faker.Faker('en_US')

    def next(self):
        _data = {}

        for attr in PersonProfile._fields.keys():
            _data[attr] = getattr(self.dataset, attr)()
            if isinstance(_data[attr], Decimal):
                _data[attr] = float(_data[attr])

        return PersonProfile(**_data)

    @staticmethod
    def schema():
        return JsonSchema(PersonProfile)


class PersonProfile(Record):
    first_name_nonbinary = String()
    last_name_nonbinary = String()
    address = String()
    ascii_email = String()
    user_name = String()
    language_name = String()
    latitude = Float()
    longitude = Float()
    ssn = String()
    ipv4 = String()
    unix_time = Integer()



from pandioml.function import FunctionBase
from pandioml.core import Pipeline, Pipelines
from pandioml.core.artifacts import artifact
from pandioml.model import LogisticRegression
from pandioml.model import StandardScaler
from pandioml.data.record import Record, Null, Boolean, Integer, Long, Float, Double, Bytes, String, Array, Map
from pandioml.metrics import Accuracy
from pandioml.model import ModelUtility


class Output(Record):
    prediction = Boolean()


class Function(FunctionBase):
    model = artifact.add('LogisticRegression_model',
                         ModelUtility.load_or_instantiate('LogisticRegression_model.pickle', LogisticRegression))
    metric = Accuracy()  #Optional
    scaler = StandardScaler()  #Optional

    def feature_extraction(self, result={}):
        result['features'] = {'feature_one': 0, 'feature_two': 1}

        return result

    # Optional
    def scale(self, result={}):
        result['features'] = self.scaler.learn_one(result['features']).transform_one(result['features'])
        return result

    def label_extraction(self, result={}):
        result['labels'] = 0

        return result

    def done(self, result={}):
        # Optional
        self.metric = artifact.add('Accuracy_Metric', self.metric.update(result['labels'], result['prediction']))

        return Output(prediction=result['prediction'])

    def pipelines(self, *args, **kwargs):
        return Pipelines().add(
            'inference',
            Pipeline(*args, **kwargs)
                .then(self.feature_extraction)
                .then(self.scale)  # Optional
                .then(self.label_extraction)
                .then(self.fit)
                .final(self.predict)
                .done(self.done)
                .catch(self.error)
        )



import logging
import sys
name = 'pandiocli'  # replace with value of "name" in setup.py


class MyFormatter(logging.Formatter):
    """
    A custom formatter that allows you to specify custom formatting options for the different
    logging levels. If no custom format is specified, it will use the 'base format'

    based on https://stackoverflow.com/a/14859558
    """
    # the base format, used when a format for a specific level is not defined
    base_format = '%(levelname)s: %(asctime)-8s [%(filename)s:%(lineno)d]: %(message)s \n'

    # defining the critical format
    critical_format = base_format

    def __init__(self):
        """
        Sets the base format according to the class variable 'base_format'
        """
        super().__init__(fmt=MyFormatter.base_format, datefmt=None, style='%')

    def format(self, record):
        """
        sets the appropriate format for each level
        See https://docs.python.org/3/library/logging.html#logging.Formatter.format
        """

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        format_orig = self._style._fmt

        # Replace the original format with one customized by logging level
        if record.levelno == logging.CRITICAL:
            self._style._fmt = MyFormatter.critical_format

        # Call the original formatter class to do the grunt work
        result = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = format_orig

        return result


# create a formatter
fmt = MyFormatter()

# code for logging to console
hdlr_console = logging.StreamHandler(sys.stdout)
hdlr_console.setFormatter(fmt)
logging.root.addHandler(hdlr_console)

'''
# Code for logging to file
hdlr_file = logging.FileHandler('spam.log')
hdlr_file.setFormatter(fmt)
logging.root.addHandler(hdlr_file)
'''



from pandioml.function import Context
from pandioml.core.artifacts import artifact
import signal
import argparse
import tracemalloc
import wrapper as wr
import os, sys

shutdown = False
tracemalloc.start(10)


def run(dataset_name, loops):
    import time
    try:
        if os.path.exists(dataset_name+'/dataset.py'):
            sys.path.insert(1, os.path.join(os.getcwd(), dataset_name))
            _dataset = __import__('dataset')
            generator = _dataset.Dataset()
        else:
            generator = getattr(__import__('pandioml.data', fromlist=[dataset_name]), dataset_name)()
    except Exception as e:
        raise Exception(f"Could not find the dataset specified at ({dataset_name}): {e}")

    w = wr.Wrapper(dataset_name=dataset_name)

    index = 0
    while True:
        start = time.time()
        c = Context()

        if shutdown or (index >= loops and loops != -1):
            break

        event = generator.next()

        print('event')
        print(event)

        result = w.process(generator.schema().encode(event).decode('UTF-8'), c)

        print('result')
        print(result)

        print('result')
        print(w.result)

        if w.result is not None:

            print(f"Actual: {int(w.result['labels'])}")
            print(f"Prediction: {int(w.result['prediction'])}")
            print("")

        end = time.time()
        print(f"Runtime ({index}) of the program is {round(end - start, 3)}")

        w.output = None

        index = artifact.add('dataset_index', (index + 1))

    w.fnc.shutdown()


def shutdown_callback(signalNumber, frame):
    global shutdown
    shutdown = True


signal.signal(signal.SIGINT, shutdown_callback)
signal.signal(signal.SIGTERM, shutdown_callback)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Test your PandioML project')
    parser.add_argument('--dataset_name', type=str, help='The name of the data set inside of pandioml.data')
    parser.add_argument('--loops', type=str, help='The number of events to process before finishing the test.',
                        required=False)

    args = parser.parse_args()
    loops = -1
    if 'loops' in args:
        loops = int(args.loops)

    run(args.dataset_name, loops)

    snapshot = tracemalloc.take_snapshot()
    top_stats = snapshot.statistics('lineno')[:10]
    for stat in top_stats:
        print(stat)



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, String, Integer, Double
import mysql.connector


class Dataset(Stream):
    """
    MySQL template showing how to connect to a database and stream the records in.

    Each record is an instance of Transaction

    class Transaction(Record):
        id = String()
        amount = Double()
        timestamp = Integer()
    """
    conn = None
    dataset = None

    def __init__(self):
        self.conn = mysql.connector.connect(
          host="localhost",
          user="yourusername",
          password="yourpassword",
          database="mydatabase"
        )
        self.dataset = self.conn.cursor()
        self.dataset.execute('SELECT * FROM transactions')

    def next(self):
        row = self.dataset.fetchone()
        return Transaction(id=row[1], amount=row[2], timestamp=row[3])

    @staticmethod
    def schema():
        return JsonSchema(Transaction)


class Transaction(Record):
    id = String()
    amount = Double()
    timestamp = Integer()



import logging, os, zipfile, hashlib, sys
from .config import Conf
import json
from shutil import copyfile
from pulsar import ConsumerType
import requests
from requests_toolbelt.multipart.encoder import MultipartEncoder
from appdirs import user_config_dir
dirname = os.path.dirname(os.path.realpath(__file__))

config = Conf()
if os.path.exists(user_config_dir('PandioCLI', 'Pandio')+'/config.json'):
    config.load(user_config_dir('PandioCLI', 'Pandio')+'/config.json')


def start(args):
    print(args)
    if args.command == 'upload':
        if 'project_folder' in args:
            root_path = os.getcwd()
            tmp_path = '/tmp/'
            path = root_path + '/' + args.project_folder
            if os.path.isdir(path):
                if os.path.exists(f"{path}/config.py"):
                    if not os.path.exists(f"{path}/deps"):
                        os.makedirs(f"{path}/deps")
                    sys.path.append(path)
                    project_config = __import__('config')

                    # Only one pipeline is supported for now
                    fnc = __import__('function')
                    pipeline_key = fnc.Function().pipelines().get_keys()[0]


                    copyfile(os.path.join(dirname, 'assets/pandioml-1.0.2-py3-none-any.whl'),
                             os.path.join(path, 'deps/pandioml-1.0.2-py3-none-any.whl'))

                    os.system(f"pip download \
                                --only-binary :all: \
                                --platform manylinux1_x86_64 \
                                --python-version 37 -r {dirname}/assets/pandioml_requirements.txt -d {path}/deps")

                    os.system(f"pip download \
                                --only-binary :all: \
                                --platform manylinux1_x86_64 \
                                --python-version 37 -r {path}/requirements.txt -d {path}/deps")

                    hash = hashlib.md5(bytes(args.project_folder, 'utf-8'))
                    tmp_path = tmp_path + hash.hexdigest() + '/'
                    if not os.path.exists(tmp_path):
                        os.makedirs(tmp_path)
                    tmp_file = args.project_folder.split('/')[-1:][0] + '.zip'
                    zipf = zipfile.ZipFile(tmp_path + tmp_file, 'w', zipfile.ZIP_DEFLATED)
                    zipf.write(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'assets/wrapper.py'), 'wrapper.py')
                    zipdir(path, zipf, args.project_folder)
                    zipf.close()
                    print(f"File located at {tmp_path}{tmp_file}")

                    arr = {
                        "name": project_config.pandio['FUNCTION_NAME'],
                        "userConfig": {
                            "pipeline": pipeline_key
                        },
                        "inputs": project_config.pandio['INPUT_TOPICS'],
                        "parallelism": 1,
                        "log-topic": project_config.pandio['LOG_TOPIC'],
                        "className": 'wrapper.Wrapper',
                        "py": tmp_file,
                        #"consumerType": ConsumerType.Shared,
                        "runtime": "PYTHON"
                    }

                    mp_encoder = MultipartEncoder(
                        fields={
                            'functionConfig': ('functionConfig', json.dumps(arr), 'application/json'),
                            'data': (tmp_file, open(tmp_path + tmp_file, 'rb'), 'application/octet-stream')
                        }
                    )

                    headers = {'Content-Type': mp_encoder.content_type}

                    _token = getattr(config, 'PANDIO_CLUSTER_TOKEN')
                    if _token is not None:
                        headers['Authorization'] = f"Bearer {_token}"

                    if 'ADMIN_API' in project_config.pandio:
                        cluster = project_config.pandio['ADMIN_API']
                    else:
                        cluster = getattr(config, 'PANDIO_CLUSTER')

                    if 'TENANT' in project_config.pandio:
                        tenant = project_config.pandio['TENANT']
                    else:
                        tenant = getattr(config, 'PANDIO_TENANT')

                    if 'NAMEPSACE' in project_config.pandio:
                        namespace = project_config.pandio['NAMESPACE']
                    else:
                        namespace = getattr(config, 'PANDIO_NAMESPACE')

                    response = requests.post(
                        f"{cluster}/admin/v3/functions/{tenant}/{namespace}/{project_config.pandio['FUNCTION_NAME']}",
                        data=mp_encoder,  # The MultipartEncoder is posted as data, don't use files=...!
                        # The MultipartEncoder provides the content-type header with the boundary:
                        headers=headers
                    )

                    if os.path.exists(tmp_path + tmp_file):
                        os.remove(tmp_path + tmp_file)

                    if response.status_code == 204:
                        print("Function uploaded successfully!")
                    else:
                        raise Exception(f"The function could not be uploaded: {response.text}")
                else:
                    raise Exception(f"File ({path}/config.py) could not be found.")
            else:
                raise Exception(f"Path ({path}) could not be found.")
        else:
            raise Exception('--folder-name must be specified')
    elif args.command == 'generate':
        if 'project_name' not in args:
            raise Exception("--project_name must be specified.")

        if not os.path.exists(args.project_name):
            logging.debug(f"Creating folder {args.project_name}")
            try:
                os.makedirs(args.project_name)
            except:
                raise Exception("Could not create folder for the project: {args.project_name}")

        copyfile(os.path.join(dirname, 'assets/function_template.py'), f"{args.project_name}/function.py")
        copyfile(os.path.join(dirname, 'assets/config_template.py'), f"{args.project_name}/config.py")
        copyfile(os.path.join(dirname, 'assets/requirements_template.txt'), f"{args.project_name}/requirements.txt")

        print(f"New function project created in: `{args.project_name}`")
        print("")
        print(f"Open {args.project_name}/function.py to begin defining your model.")
        print("")
        print(f"For help creating your function, see existing examples or read the documentation in the README.")
        print("")
    else:
        raise Exception('Nothing matched the action: {0}'.format(args.command))


def zipdir(path, ziph, project_folder):
    # ziph is zipfile handle
    for root, dirs, files in os.walk(path):
        for file in files:
            if 'test' not in file and 'runner.py' not in file:
                rel_dir = os.path.relpath(root, path)
                if rel_dir == '.' or rel_dir == 'deps':
                    project_folder = str(project_folder).split("/")[-1:][0]
                    if rel_dir == '.':
                        rel_dir = f"{project_folder}/src"
                    if rel_dir == 'deps':
                        rel_dir = f"{project_folder}/deps"
                    rel_file = os.path.join(rel_dir, file)
                    ziph.write(os.path.join(root, file), rel_file)



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, String, Integer, Double
import csv
import os


class Dataset(Stream):
    """
    CSV template showing how to load a CSV and stream the records in.

    Each record is an instance of Transaction

    class Transaction(Record):
        id = String()
        amount = Double()
        timestamp = Integer()
    """
    dataset = None
    path = os.path.dirname(os.path.realpath(__file__)) + '/data.csv'

    def __init__(self):
        self.dataset = iter(self._reader())

    def _reader(self):
        with open(self.path, 'r') as data:
            reader = csv.reader(data)
            for row in reader:
                yield row

    def next(self):
        row = next(self.dataset)
        return Transaction(id=row[0], amount=float(row[1]), timestamp=int(row[2]))

    @staticmethod
    def schema():
        return JsonSchema(Transaction)


class Transaction(Record):
    id = String()
    amount = Double()
    timestamp = Integer()



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, String, Integer, Double


class Dataset(Stream):
    """
    Dataset template showing how to stream the records in.

    Each record is an instance of Transaction

    class Transaction(Record):
        id = String()
        amount = Double()
        timestamp = Integer()
    """
    dataset = None

    def __init__(self):
        self.dataset = iter([{
            'id': '1',
            'amount': 2.50,
            'timestamp': 123456789
        }])

    def next(self):
        return Transaction(**next(self.dataset))

    @staticmethod
    def schema():
        return JsonSchema(Transaction)


class Transaction(Record):
    id = String()
    amount = Double()
    timestamp = Integer()



import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from pandioml.function import Function
import dataset as _dataset
import config


class Wrapper(Function):
    _dataset = None

    def __init__(self):
        pass

    def process(self, input, context):
        if 'OUTPUT_TOPICS' in config.pandio:
            d = _dataset.Dataset()
            while True:
                try:
                    for output_topic in config.pandio['OUTPUT_TOPICS']:
                        context.publish(output_topic, _dataset.Dataset.schema().encode(d.next()).decode('UTF-8'))
                except StopIteration:
                    break
        else:
            print('No output topics defined.')



import logging
import argparse
import sys
import os
from appdirs import user_config_dir

if not os.path.exists(user_config_dir('PandioCLI', 'Pandio')):
    logging.debug(f"Creating folder {user_config_dir('PandioCLI', 'Pandio')}")
    try:
        os.makedirs(user_config_dir('PandioCLI', 'Pandio'))
    except:
        raise Exception("Could not create folder for the project: {args.project_name}")

from src import function, register, config, test, dataset

welcome_text = """
                                                                 dddddddd                         
PPPPPPPPPPPPPPPPP                                                d::::::d  iiii                   
P::::::::::::::::P                                               d::::::d i::::i                  
P::::::PPPPPP:::::P                                              d::::::d  iiii                   
PP:::::P     P:::::P                                             d:::::d                          
  P::::P     P:::::Paaaaaaaaaaaaa  nnnn  nnnnnnnn        ddddddddd:::::d iiiiiii    ooooooooooo   
  P::::P     P:::::Pa::::::::::::a n:::nn::::::::nn    dd::::::::::::::d i:::::i  oo:::::::::::oo 
  P::::PPPPPP:::::P aaaaaaaaa:::::an::::::::::::::nn  d::::::::::::::::d  i::::i o:::::::::::::::o
  P:::::::::::::PP           a::::ann:::::::::::::::nd:::::::ddddd:::::d  i::::i o:::::ooooo:::::o
  P::::PPPPPPPPP      aaaaaaa:::::a  n:::::nnnn:::::nd::::::d    d:::::d  i::::i o::::o     o::::o
  P::::P            aa::::::::::::a  n::::n    n::::nd:::::d     d:::::d  i::::i o::::o     o::::o
  P::::P           a::::aaaa::::::a  n::::n    n::::nd:::::d     d:::::d  i::::i o::::o     o::::o
  P::::P          a::::a    a:::::a  n::::n    n::::nd:::::d     d:::::d  i::::i o::::o     o::::o
PP::::::PP        a::::a    a:::::a  n::::n    n::::nd::::::ddddd::::::ddi::::::io:::::ooooo:::::o
P::::::::P        a:::::aaaa::::::a  n::::n    n::::n d:::::::::::::::::di::::::io:::::::::::::::o
P::::::::P         a::::::::::aa:::a n::::n    n::::n  d:::::::::ddd::::di::::::i oo:::::::::::oo 
PPPPPPPPPP          aaaaaaaaaa  aaaa nnnnnn    nnnnnn   ddddddddd   dddddiiiiiiii   ooooooooooo   
"""


def main():
    logging.root.setLevel(logging.DEBUG)
    args = parse_cmd_args(sys.argv[1:])
    # if verbose flag not passed on as an argument, this will disable all logging levels
    if not args.verbose:
        logging.disable(logging.CRITICAL)  # This will disable all logging
        # the table below shows the logging levels and there value. if you set the level above (changing "CRITICAL") will
        # change what message is shown. any logging types with a numeric value BELOW AND EQUAL the set level will not be show
        # e.g. if the level is ERROR, only CRITICAL will be shown as 50 > 40

        # | Level    | Value |
        # |----------|-------|
        # | CRITICAL | 50    |
        # | ERROR    | 40    |
        # | WARNING  | 30    |
        # | INFO     | 20    |
        # | DEBUG    | 10    |
        # | NOTSET   | 0     |

    # if quiet flag is enabled, stdout (console output) is written to devnull where data is discarded
    if args.quiet:
        sys.stdout = open(os.devnull, 'a')

    #########################################
    #        Application code below         #
    #########################################

    # # prints all the arguments as Namespace object https://docs.python.org/3/library/argparse.html#the-namespace-object
    # print("args -->", args)

    # # print all the arguments as a dictionary
    # print('vars(args) -->', vars(args))

    # identifying if a subparser is invoked. if invoked, call appripriate function
    if 'func' in vars(args):
        logging.debug('calling the appropriate function for parser')
        args.func(args)

    # # below are logging levels with "debug" being the lowest and "critical" being the highest
    # logging.debug(
    #     'The lowest level. Used for small details. Usually you care about these messages only when diagnosing problems.')
    # logging.info('Used to record information on general events in your program or confirm that things are working at their point in the program.')
    # logging.warning(
    #     'Used to indicate a potential problem that doesn’t prevent the program from working but might do so in the future.')
    # logging.error(
    #     'Used to record an error that caused the program to fail to do something')
    # logging.critical(
    #     'The highest level. Used to indicate a fatal error that has caused or is about to cause the program to stop running entirely.')


def parse_cmd_args(cmd_args):

    # refactoring so argparse is testable available at https://stackoverflow.com/a/18161115
    #####################################################################
    #            Code for parsing command line arguments                #
    #####################################################################

    # https://docs.python.org/3.7/howto/argparse.html
    # https://docs.python.org/3/library/argparse.html

    print(welcome_text)

    # create the arguent parser
    parser = argparse.ArgumentParser(
        description="Pandio's Machine Learning Service")

    # mutually exclusive means that only one option can be supplied. supplying both will result in an error
    group = parser.add_mutually_exclusive_group()

    # "store_true" means that if the argument is provided, the value of the argument is true
    group.add_argument('-v', '--verbose', action='store_true',
                       help='Give more output')
    group.add_argument('-q', '--quiet', action='store_true',
                       help='Give no output')

    # the following two blocks of code are commented out as they are not relevent to the example but remain
    # to demonstrate how to write arguments

    # # a mandatory argument. we also specified that only integers can be supplied
    # parser.add_argument(
    #     'square', help='This is the help for echo, a mandatory argument of type int', type=int)

    # # an argument that accepts multiple value seperated by a comma. the "-" means it is optional
    # parser.add_argument(
    #     '--multiple', help='This optional option accepts multiple options', nargs='*')

    # code for subparsers
    # dest='subparser_name' is used to identify the subparser name
    subparsers = parser.add_subparsers(
        title='List of sub commands', description='A description of all the available sub commmands', help='All the commands', dest='subparser_name')

    # code for subparser command r
    parser_r = subparsers.add_parser('register', help='register for a Pandio account')
    parser_r.add_argument('email', type=str, help='email to register from')
    # a function to call when subparser invoked
    parser_r.set_defaults(func=register.start)

    # code for subparser command c
    parser_c = subparsers.add_parser('config', help='Update the configuration for the pandiocli')
    parser_c.add_argument('command', type=str, help='command for configuration management')
    parser_c.add_argument('--key', type=str, help='The key to set', required=False)
    parser_c.add_argument('--value', type=str, help='The value to set to the key', required=False)
    # a function to call when subparser invoked
    parser_c.set_defaults(func=config.start)

    # code for subparser command t
    parser_t = subparsers.add_parser('test', help='Test your PandioML project')
    parser_t.add_argument('--project_folder', type=str, help='The path to the project to test.')
    parser_t.add_argument('--dataset_name', type=str, help='The name of the data set inside of pandioml.data')
    parser_t.add_argument('--loops', type=str, help='The number of events to process before finishing the test.',
                          required=False)
    parser_t.add_argument('--workers', type=str, help='The number of script iterations to create. This adds parallelism'
                                                      ' to the execution of the function.', required=False)
    # a function to call when subparser invoked
    parser_t.set_defaults(func=test.start)

    # create the parser for the "function" command
    parser_f = subparsers.add_parser('function', help='manage functions')
    parser_f.add_argument('command', type=str, help='command for function management')
    parser_f.add_argument('--project_folder', type=str, required=False,
                          help='function folder name for the relevant command and action')
    parser_f.add_argument('--project_name', type=str, required=False,
                          help='folder name for the dataset project')
    parser_f.add_argument('--type', type=str, required=False,
                          help='type of dataset project to generate')
    parser_f.set_defaults(func=function.start)

    # create the parser for the "dataset" command
    parser_d = subparsers.add_parser('dataset', help='manage datasets')
    parser_d.add_argument('command', type=str, help='command for dataset management')
    parser_d.add_argument('--project_folder', type=str, required=False,
                          help='function folder name for the relevant command and action')
    parser_d.add_argument('--project_name', type=str, required=False,
                          help='folder name for the dataset project')
    parser_d.add_argument('--type', type=str, required=False,
                          help='type of dataset project to generate')
    parser_d.set_defaults(func=dataset.start)

    # if no arguments are given i.e. only the command name is invoked. this will ensure that the help message is printed out
    if len(cmd_args) == 0:
        parser.print_help()
        sys.exit(1)

    # writing the arguments to a variable to be accessed
    parsed = parser.parse_args(cmd_args)
    return parsed


if __name__ == "__main__":
    main()



import pathlib
import os
import os.path
import signal
import subprocess
import sys
from .config import Conf
import hashlib
from pandioml.core.artifacts import artifact
import zipfile
from appdirs import user_config_dir

config = Conf()
if os.path.exists(user_config_dir('PandioCLI', 'Pandio')+'/config.json'):
    config.load(user_config_dir('PandioCLI', 'Pandio')+'/config.json')

shutdown = False


def start(args):
    BUF_SIZE = 65536

    loops = -1
    if args.loops is not None:
        loops = int(args.loops)

    workers = 1
    if args.workers is not None:
        workers = int(args.workers)

    if not os.path.isfile(f"{args.project_folder}/function.py"):
        print(f"{args.project_folder} is invalid.")
        exit()

    val = input("Would you like to store artifacts (dataset, pipeline, model, etc.)? If no, only the model will be saved. (y,n): ")
    store = True if val.lower() == 'y' or val.lower() == 'yes' else False
    if store:
        print("Artifacts will be saved!")

        print("")

        name_id = input("Would you like to name this run? Leave blank for no name: ")

        print("")

        print("Generating artifact identifiers.")

        os.system(
            f"cd {args.project_folder} && rm -rf dist && rm -rf build && rm -rf function.spec && rm -rf __pycache__ && "
            f"PYTHONHASHSEED=1 && export "
            "PYTHONHASHSEED && pyinstaller -F --exclude-module scikit-learn "
            "--exclude-module river --exclude-module pandioml --exclude-module scikit-multiflow --exclude-module "
            "pulsar-client --exclude-module Faker --exclude-module scipy --exclude-module numpy --exclude-module "
            "pandas --exclude-module python-dateutil --exclude-module six --exclude-module test-unidecode "
            "--exclude-module ratelimit --exclude-module fastavro --exclude-module grpcio --exclude-module certifi "
            "--exclude-module prometheus-client --exclude-module apache-bookkeeper-client --exclude-module protobuf "
            "--exclude-module pytz --exclude-module sortedcontainers --exclude-module matplotlib --exclude-module "
            "requests --exclude-module pymmh3 --exclude-module setuptools --exclude-module pyparsing --exclude-module"
            " pillow --exclude-module cycler --exclude-module kiwisolver --exclude-module idna --exclude-module "
            "chardet --exclude-module urllib3 --exclude-module threadpoolctl --exclude-module sklearn "
            "--exclude-module pytest --exclude-module pickle function.py >/dev/null 2>&1 && unset PYTHONHASHSEED")

        md5 = hashlib.md5()

        with open(f"{args.project_folder}/dist/function", 'rb') as f:
            while True:
                data = f.read(BUF_SIZE)
                if not data:
                    break
                md5.update(data)

        os.system(f"cd {args.project_folder} && rm -rf dist && rm -rf build && rm -rf function.spec && rm -rf __pycache__")

        if len(name_id) > 0:
            artifact.set_name_id(name_id)

        artifact.set_pipeline_id(md5.hexdigest())

        print(f"Artifact pipeline id is: {artifact.get_pipeline_id()}")

        print(f"Artifact name id is: {artifact.get_name_id()}")
    else:
        print("NOT SAVING any artifacts!")

    body = ''
    with open(f"{args.project_folder}/function.py", 'r') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            body += data

    def zipit(storage_location):
        def zipdir(path, ziph):
            # ziph is zipfile handle
            for root, dirs, files in os.walk(path):
                for file in files:
                    rel_dir = os.path.relpath(root, path)
                    if 'test' not in file and 'runner.py' not in file and \
                            'artifacts' not in rel_dir and 'pycache' not in rel_dir:
                        rel_file = os.path.join(rel_dir, file)
                        ziph.write(os.path.join(root, file), rel_file)

        zipf = zipfile.ZipFile(storage_location + '/pipeline_files.zip', 'w', zipfile.ZIP_DEFLATED)
        zipdir(args.project_folder, zipf)
        zipf.close()

    artifact.add('archive_pipeline', zipit)

    print("Starting execution of pipeline(s).")

    sys.path.insert(1, os.path.join(os.path.abspath(os.path.dirname(__file__)), 'assets'))
    sys.path.insert(1, os.path.join(os.getcwd(), args.project_folder))
    pm = __import__('runner')
    pm.run(args.dataset_name, loops)

    print("")


def shutdown_callback(signalNumber, frame):
    global shutdown
    shutdown = True


signal.signal(signal.SIGINT, shutdown_callback)
signal.signal(signal.SIGTERM, shutdown_callback)



import os

pandio = {
    'FUNCTION_NAME': 'exampleFunction123',
    'CONNECTION_STRING': 'pulsar://localhost:6651',
    'INPUT_TOPICS': ['non-persistent://public/default/in'],
    'OUTPUT_TOPICS': ['non-persistent://public/default/out'],
    'LOG_TOPIC': 'non-persistent://public/default/log',
    'ARTIFACT_STORAGE': f"{os.path.dirname(os.path.realpath(__file__))}/artifacts"
}



import setuptools, os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="pandiocli",
    version="1.0.8",
    author="Joshua Odmark",
    author_email="josh@pandio.com",
    description="CLI to control Pandio's machine learning service.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/pandio-com/pandiocli",  # usually to github repository
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Documentation",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
    ],
    entry_points={
        "console_scripts": [
            'pandiocli = src.__main__:main'
        ]
    },
    install_requires=['goodconf==1.0.0', 'requests==2.25.1', 'pandioml==1.0.4', 'pyinstaller==4.3', 'Faker==8.1.1',
                      'appdirs==1.4.4', 'requests-toolbelt==0.9.1'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests'
)


import os
from goodconf import GoodConf, Value
from appdirs import user_config_dir


class PandioConf(GoodConf):
    @classmethod
    def set_value(cls, key, value):
        if hasattr(cls, key):
            return setattr(cls, key, value)

        return False


class Conf(PandioConf):
    "Configuration for pandiocli"
    DEBUG = Value(default=False, help="Toggle debugging.")
    PANDIO_CLUSTER = Value(
        default='NOT_CONFIGURED',
        help="Pandio cluster url")
    PANDIO_TENANT = Value(
        default='NOT_CONFIGURED',
        help="Pandio cluster tenant")
    PANDIO_NAMESPACE = Value(
        default='NOT_CONFIGURED',
        help="Pandio cluster namespace")
    PANDIO_CLUSTER_TOKEN = Value(
        default='NOT_CONFIGURED',
        help="Pandio cluster authorization token")
    PANDIO_DATA_TOKEN = Value(
        default='NOT_CONFIGURED',
        help="Pandio data authorization token")
    PANDIO_REGISTER_TOKEN = Value(
        default=False,
        help="Pandio register authorization token")
    PANDIO_EMAIL = Value(
        default='NOT_CONFIGURED',
        help="Pandio account email")


config = Conf()
if os.path.exists(user_config_dir('PandioCLI', 'Pandio')+'/config.json'):
    config.load(user_config_dir('PandioCLI', 'Pandio')+'/config.json')


def start(args):
    if args.command == 'set' and 'key' in args and 'value' in args:
        for k in config._values:
            config.set_value(k, getattr(config, k))
        if config.set_value(args.key, args.value) is not False:
            print(f"Setting {args.key} to {args.value}")
            f = open(user_config_dir('PandioCLI', 'Pandio') + '/config.json', "w+")
            f.write(config.generate_json())
            f.close()
        else:
            print(f"Could not set {args.key} to {args.value}, {args.key} does not exist.")

    elif args.command == 'show':
        print("PANDIO SETTINGS")
        print("")
        for k in config._values:
            print(k, '=', getattr(config, k))
    elif args.command == 'file':
        print("PANDIO SETTINGS FILE LOCATION")
        print("")
        print(user_config_dir('PandioCLI', 'Pandio') + '/config.json')
    elif args.command == 'reset':
        print("RESET PANDIO SETTINGS")
        print("")
        os.remove(user_config_dir('PandioCLI', 'Pandio') + '/config.json')
        print('Deleted file: ' + user_config_dir('PandioCLI', 'Pandio') + '/config.json')
    else:
        print(f"Action ({args.action}) not found.")

    print("")



import sys
import os
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from pandioml.function import Function
from pandioml.core import Pipelines
import function as pm
import config
from pandioml.core.artifacts import artifact
import time
from pandioml.data.record import JsonSchema

artifact.set_storage_location(config.pandio['ARTIFACT_STORAGE'])


class Wrapper(Function):
    fnc = None
    result = None
    input_schema = None

    def __init__(self, dataset_name=None):
        artifact.add('runtime_settings', {'config.pandio': config.pandio, 'sys.version': sys.version,
                                          'timestamp': time.strftime("%Y%m%d-%H%M%S")})
        if dataset_name is not None:
            try:
                if os.path.exists(dataset_name + '/dataset.py'):
                    sys.path.insert(1, os.path.join(os.getcwd(), dataset_name))
                    _dataset = __import__('dataset')
                    self.input_schema = _dataset.Dataset.schema()
                else:
                    self.input_schema = getattr(__import__('pandioml.data', fromlist=[dataset_name]),
                                                dataset_name).schema()
            except Exception as e:
                raise Exception(f"Could not find the dataset specified at ({dataset_name}): {e}")

    def process(self, input, context):
        self.fnc = pm.Function(self.input_schema.decode(input), context, config)
        try:
            self.fnc.startup()
        except Exception as e:
            raise Exception(f"Could not execute startup method: {e}")

        try:
            p = self.fnc.pipelines()
        except Exception as e:
            raise Exception(f"Could not build pipelines: {e}")

        if isinstance(p, Pipelines) is False:
            raise Exception(f"Method pipelines should return a Pipelines object!")

        # Set the first pipeline, multiple are not currently supported
        context.set_user_config_value('pipeline', p.get_keys()[0])

        output = p.go(context.get_user_config_value('pipeline'), self.fnc)
        if isinstance(output[context.get_user_config_value('pipeline')], tuple) and isinstance(output[context.get_user_config_value('pipeline')][0], Exception):
            raise Exception(f"An exception occurred in the pipeline: {output[context.get_user_config_value('pipeline')][0]} {output[context.get_user_config_value('pipeline')][1]}")
            # TODO, see if this should be a print and continue, or halt the entire execution

        self.result = self.fnc.get_result()

        if 'OUTPUT_TOPICS' in config.pandio:
            for output_topic in config.pandio['OUTPUT_TOPICS']:
                if output is not None:
                    context.publish(output_topic, JsonSchema(getattr(pm, output[context.get_user_config_value('pipeline')].schema()['name'])).encode(output[context.get_user_config_value('pipeline')]).decode('UTF-8'))
                else:
                    print("Warning, output variable is empty, should be defined in self.fnc.done method.")

        if artifact.get_name_id() is not None:
            context.incr_counter(artifact.get_name_id(), 1)

            count = context.get_counter(artifact.get_name_id())

            if count > 0 and count % 1000 == 0:
                #self.fnc.sync_models(context)
                if artifact.get_pipeline_id() is not None:
                    artifact.save(checkpoint=True)

        return input



import logging, pathlib, os
from .config import Conf
import requests
import faker
import time
from appdirs import user_config_dir

config = Conf()
if os.path.exists(user_config_dir('PandioCLI', 'Pandio')+'/config.json'):
    config.load(user_config_dir('PandioCLI', 'Pandio')+'/config.json')


def save_config():
    f = open(user_config_dir('PandioCLI', 'Pandio') + '/config.json', "w")
    f.write(config.generate_json())
    f.close()


def register(email):
    fake = faker.Faker('en_US')
    response = requests.post(
        "https://app.pandio.com/api/v1/starter/create-user",
        json={'email': email, 'first_name': getattr(fake, 'first_name_nonbinary')(),
              'last_name': getattr(fake, 'last_name_nonbinary')()}
    )

    if response.status_code != 201:
        return None

    return response.json()['token']


def get_details(token):
    response = requests.get(
        "https://app.pandio.com/api/v1/starter/tenant-details",
        headers={'Authorization': f"Bearer {token}", 'Accept': 'application/json'}
    )

    if response.status_code != 200:
        print(response.json()['error'])
        exit()

    if response.json()['status'] != 'HEALTH_OK':
        print("")
        print("Retrieving...")
        return None

    return response.json()


def start(args):
    print(f"Creating an account with email {args.email}.")
    print("")
    existing_token = getattr(config, 'PANDIO_REGISTER_TOKEN')
    if existing_token is False:
        token = register(args.email)
        if token is None:
            print("")
            print("An error occurred. Please try again.")
            exit()

        config.set_value('PANDIO_REGISTER_TOKEN', token)

        save_config()

        input("Please verify your email and then return to this prompt and hit enter:")

        print("")
    else:
        token = existing_token

    print("Retrieving your account details.")

    time.sleep(10)

    for i in range(0, 3):
        details = get_details(token)
        if details is None:
            time.sleep(20)
            continue
        else:
            break

    if details is None:
        print("")
        print("Something went wrong with account activation. "
              "Please try again or reach out to customer support at support@pandio.com.")
        exit()

    config.set_value('PANDIO_CLUSTER', details['data']['cluster'])
    config.set_value('PANDIO_TENANT', details['data']['tenant'])
    config.set_value('PANDIO_NAMESPACE', details['data']['namespace'])
    config.set_value('PANDIO_CLUSTER_TOKEN', details['data']['cluster_token'])
    config.set_value('PANDIO_DATA_TOKEN', details['data']['training_data_token'])
    config.set_value('PANDIO_EMAIL', args.email)

    print("")
    print("Your account is now active, cheers!")
    print("")

    save_config()
